#!/usr/bin/perl

use strict;
use warnings;

use constant { 
    SCOO_EXEC => "/home/vu/workspace/cuda_spmv_scoo/cuda_bwc",
    PREP_EXEC => "/home/vu/workspace/cuda_spmv_scoo/prep",
    CUSP_EXEC => "/home/vu/workspace/cusp-bench/benchcusp",
    NTIMES => 1000,
    BENCH => { 'scoo' => 1, 'cusp' => 0, 'reprep' => 1,
        'blocksize' => 64, 
        'gpu' => 1,
    }
};

my $matrix_file = shift;
my $gpu = shift;
my $blocksize = shift;

$gpu = BENCH->{'gpu'} if (!$gpu);
$blocksize = BENCH->{'blocksize'} if (!$blocksize);

die("Please input matrix name") if (!$matrix_file);

print "start benching ... " . $matrix_file . "\n";

die("Can't find $matrix_file") if (!-f $matrix_file);

if (BENCH->{'scoo'} == 1) {
    print"PREP...\n";
    if (! -f "$matrix_file.prep.$blocksize" || BENCH->{'reprep'}) {
        system( PREP_EXEC . " -b $blocksize -m $matrix_file -g $gpu -t") == 0
            or die("$?");
        system( "touch $matrix_file.prep.$blocksize");
    }
    print "SCOO...\n";
    system( SCOO_EXEC . " -g $gpu -b $blocksize -m $matrix_file -n " . NTIMES ) == 0 
        or die("$?");
}

if (BENCH->{'cusp'} == 1) {
    print "CUSP...\n";
    system( CUSP_EXEC . " -g $gpu -b $blocksize -m $matrix_file -n ". NTIMES ) == 0
        or die("$?");
}

print "DONE!\n";
