#ifndef SLICED_COO_H_
#define SLICED_COO_H_

#include "../util/util.h"
#include <thrust/host_vector.h>
#include <thrust/device_vector.h>
#include <sstream>
#include "xl_sliced_coo.h"
#include "xl_sliced_coo_kernel.cuh"
#include "../util/textures.h"

namespace cuda_bwc{

template<typename T, const uint32_t LANE_SIZE>
class SlicedCoo : public XLSlicedCoo<T>{
public:
	SlicedCoo() : XLSlicedCoo<T>(0,0) {};
	SlicedCoo(uint32_t num_rows, uint32_t num_cols) :
		XLSlicedCoo<T>(num_rows, num_cols) {};

	std::string getCacheName();
	void multiply(const T *v, T *r);
    static const uint32_t GROUP_SIZE = (1<<30);

    uint32_t numRowsPerSlice(){return NUM_ROWS_PER_SLICE;};
    uint32_t getLaneSize() { return LANE_SIZE; }

private:
	static const uint32_t THREADS_PER_BLOCK = 1024;
//    static const uint32_t LANE_SIZE = 16;
	static const uint32_t NUM_ROWS_PER_SLICE = SHARED_MEMORY_SIZE/sizeof(T)/LANE_SIZE;

};

template <typename T, uint32_t LANE_SIZE>
std::string SlicedCoo<T, LANE_SIZE>::getCacheName(){
	std::stringstream ss;
	ss << "-" << LANE_SIZE << "scoo-" << NUM_ROWS_PER_SLICE << "pack-" << this->num_rows << "-" << this->num_cols << ".bin";
	return ss.str();
}

/*
template<>
void XLSlicedCoo<uint32_t>::multiply(const uint32_t *v, uint32_t *r){
	static const uint32_t NUM_BLOCKS = getNumMultiprocessors();
	xl_sliced_coo_kernel_32<uint32_t, THREADS_PER_BLOCK, NUM_ROWS_PER_SLICE> <<<NUM_BLOCKS, THREADS_PER_BLOCK>>>(
						this->num_rows,
		                divideAndCeil(this->num_rows, NUM_ROWS_PER_SLICE),
		                divideAndCeil(this->num_cols, GROUP_SIZE),
		                GROUP_SIZE,
		                thrust::raw_pointer_cast(&dv_coo[0]),
                        thurst::raw_pointer_cast(&dv_value[0]),
		                thrust::raw_pointer_cast(&dv_offsets[0]),
		                v,
		                r);
}*/

template<typename T, const uint32_t LANE_SIZE>
void SlicedCoo<T, LANE_SIZE>::multiply(const T *v, T *r){
//	static const uint32_t NUM_BLOCKS = getNumMultiprocessors();
    bind_x(v);
    	sliced_coo_kernel_32<T, THREADS_PER_BLOCK, NUM_ROWS_PER_SLICE, LANE_SIZE> <<<divideAndCeil(this->num_rows, NUM_ROWS_PER_SLICE), THREADS_PER_BLOCK>>>( //dim3(THREADS_PER_BLOCK/4, 4)>>>(
						this->num_rows,
						divideAndCeil(this->num_rows, NUM_ROWS_PER_SLICE),
						thrust::raw_pointer_cast(&this->dv_col[0]),
                        thrust::raw_pointer_cast(&this->dv_row[0]),
                        thrust::raw_pointer_cast(&this->dv_value[0]),
						thrust::raw_pointer_cast(&this->dv_offsets[0]),
						v,
						r);
    unbind_x(v);
}

//template<>
//void XLSlicedCoo<uint32_t>::multiply(const uint32_t *v, uint32_t *r){
//	static const uint32_t NUM_BLOCKS = getNumMultiprocessors();
//	xl_sliced_coo_kernel_32<uint32_t, THREADS_PER_BLOCK, NUM_ROWS_PER_SLICE> <<<NUM_BLOCKS, THREADS_PER_BLOCK>>>(
//					this->num_rows,
//	                divideAndCeil(this->num_rows, NUM_ROWS_PER_SLICE),
//	                divideAndCeil(this->num_cols, GROUP_SIZE),
//	                GROUP_SIZE,
//	                thrust::raw_pointer_cast(&dv_coo[0]),
//	                thrust::raw_pointer_cast(&dv_offsets[0]),
//	                v,
//	                r);
//}

}

#endif /* SLICED_COO_H_ */
