/*
 * sliced_ellpack_kernel.h
 *
 *  Created on: Dec 16, 2010
 *      Author: hans
 */

#ifndef SLICED_ELLPACK_KERNEL_H_
#define SLICED_ELLPACK_KERNEL_H_

__device__ __constant__ uint8_t LO_HI_BYTE_MASK[2] = {0x0F, 0xF0};

__inline__ __host__ __device__ uint32_t get_elem(const uint8_t MASK[], const uint8_t * Aj, const uint32_t at){
    uint32_t pos = at*5/2;
    uint32_t temp = 0;
    const uint32_t bit = at & 1;
    const int32_t add = 1 - (bit * 2);
    const uint32_t shift = 16 - bit*4;
    pos += bit * 2;

    temp = Aj[pos];
    pos += add;
    temp |= Aj[pos] << 8;
    pos += add;
    temp |= (Aj[pos] & MASK[bit]) << shift;

    return temp;
}

__inline__ __host__ __device__ void get_border(const uint8_t *borders, uint32_t at, uint16_t& border, uint8_t& base){
    uint32_t pos = at * 3;
    border = borders[pos] | (borders[pos+1]<<8);
    base = borders[pos+2];
}

__inline__ __device__ uint32_t getElem(uint32_t at, const uint8_t* Aj){
//    const uint32_t pos = at*3;
    uint32_t rs = Aj[at];
    rs|=Aj[at+1]<<8;
    rs|=Aj[at+2]<<16;
    return rs;
}

__device__ __constant__ uint32_t invalid_index = 0;

template <typename ValueType, const uint32_t height>
__global__ void
sliced_ellpack_kernel(const uint32_t num_rows,
                const uint32_t * Ap,
                const uint32_t * Aj,
                const ValueType *V,
                const ValueType * x,
                      ValueType * y)
{
    const uint32_t thread_id = blockDim.x * blockIdx.x + threadIdx.x;
    const uint32_t grid_size = gridDim.x * blockDim.x;
    const uint32_t odd = num_rows & (height - 1);
 
    uint32_t row = thread_id;
    for(; row < num_rows - odd; row += grid_size)
    {
        ValueType sum = 0;
        const uint32_t pp = row/height;
        const uint32_t start = fetch_ap(pp,Ap)+ threadIdx.x;
        const uint32_t end   = fetch_ap(pp+1,Ap);

        for(uint32_t n = start; n < end; n+=height) {
            const uint32_t col = Aj[n]; //getElem(n,Aj);
            if (col != invalid_index)
                 sum += fetch_x(col-1, x) * V[n];
        }
        y[row] = sum;
    }
    
    for (;row < num_rows; row+=grid_size) {
        ValueType sum = 0;
        const uint32_t pp = row/height;
        const uint32_t start = fetch_ap(pp,Ap)+ threadIdx.x;
        const uint32_t end   = fetch_ap(pp+1,Ap);

        for(uint32_t n = start; n < end; n+=odd) {
            const uint32_t col = Aj[n]; //getElem(n,Aj);
            if (col != invalid_index)
                 sum += fetch_x(col-1, x) * V[n];
        }
        y[row] = sum;
    }

}


//THREADS_PER_BLOCK must be == height
template <typename ValueType, const uint32_t THREADS_PER_BLOCK>
__global__ void
sliced_ellpack_kernel_x(const uint32_t num_rows,
                const uint32_t height,
                const uint32_t * Ap,
                const uint8_t * Aj,
                const uint8_t * borders,
                const uint32_t * row_start,
                const ValueType * x,
                      ValueType * y)
{

    const uint32_t thread_id = THREADS_PER_BLOCK * blockIdx.x + threadIdx.x;
    const uint32_t grid_size = gridDim.x * THREADS_PER_BLOCK;
    const uint32_t limit = (num_rows/height)*height;
    const uint32_t lastHeight = num_rows - limit;
    const uint32_t BASE = 1 << 20;
    const uint32_t invalid_index = 0;

    uint32_t row;
    for(row = thread_id; row < limit; row += grid_size)
    {
        ValueType sum = 0;
        const uint32_t pp = row/height;
        const uint32_t slice_start = fetch_ap(pp, Ap);
        const uint32_t start = slice_start + threadIdx.x;
        const uint32_t end   = fetch_ap(pp+1, Ap);
        const uint32_t ncols = (end-slice_start)/height;
        uint32_t base_val = 0;
        uint8_t new_base_val = 0;

        uint32_t index = row_start[row];
        uint32_t endIndex = row_start[row+1];

        uint16_t borderIndex = 0xFFFF;
        if(index<endIndex){
            get_border(borders, index, borderIndex, new_base_val);
        }

        for(uint32_t n = 0; n < ncols; n++)
        {
            if(n==borderIndex){
                base_val = new_base_val * BASE;
                if(++index<endIndex){
                    get_border(borders, index, borderIndex, new_base_val);
                }
            }
            uint32_t col = get_elem(LO_HI_BYTE_MASK, Aj, start+n*height) + base_val;

            if(col != invalid_index){
                sum ^= fetch_x(col-1, x);
            }

        }

        syncthreads();
        y[row] = sum;
    }

    for(; row<num_rows; row+=grid_size){
        ValueType sum = 0;
        const uint32_t pp = row/height;
        const uint32_t slice_start = fetch_ap(pp, Ap);
        const uint32_t end   = fetch_ap(pp+1, Ap);
        const uint32_t start = slice_start + threadIdx.x;
        const uint32_t ncols = (end-slice_start)/lastHeight;

        uint32_t base_val = 0;
        uint8_t new_base_val = 0;

        uint32_t index = row_start[row];
        uint32_t endIndex = row_start[row+1];

        uint16_t borderIndex = 0xFFFF;
        if(index<endIndex){
            get_border(borders, index, borderIndex, new_base_val);
        }

        for(uint32_t n = 0; n < ncols; n++)
        {
            if(n==borderIndex){
                base_val = new_base_val * BASE;
                if(++index<endIndex){
                    get_border(borders, index, borderIndex, new_base_val);
                }
            }
            uint32_t col = get_elem(LO_HI_BYTE_MASK, Aj, start+n*lastHeight) + base_val;

            if(col != invalid_index){
                sum ^= fetch_x(col-1, x);
            }
        }
        y[row] = sum;
    }
}

#endif /* SLICED_ELLPACK_KERNEL_H_ */
