/*
 * small_sliced_coo.h
 *
 *  Created on: Dec 13, 2010
 *      Author: hans
 */

#ifndef SMALL_SLICED_COO_H_
#define SMALL_SLICED_COO_H_

#include <sstream>
#include "large_sliced_coo.h"
#include "small_sliced_coo_kernel.cuh"

namespace cuda_bwc{

template<typename T, const uint32_t THREADS_PER_BLOCK>
class SmallSlicedCoo : public LargeSlicedCoo<T>{
public:
	SmallSlicedCoo() : LargeSlicedCoo<T>(0,0){};
	SmallSlicedCoo(uint32_t num_rows, uint32_t num_cols) :
		LargeSlicedCoo<T>(num_rows, num_cols){};
	std::string getCacheName();
	void multiply(const T *v, T *r);

protected:
	uint32_t numRowsPerSlice(){return NUM_ROWS_PER_SLICE;};

private:
//	static const uint32_t THREADS_PER_BLOCK = ThreadsPerBlock; //sizeof(T)<= 4? 512:256;
	static const uint32_t NUM_ROWS_PER_SLICE = SHARED_MEMORY_SIZE/THREADS_PER_BLOCK/sizeof(T);
};

template <typename T, const uint32_t THREADS_PER_BLOCK>
std::string SmallSlicedCoo<T, THREADS_PER_BLOCK>::getCacheName(){
	std::stringstream ss;
	ss << "-sscoo-" << NUM_ROWS_PER_SLICE << "pack-" << this->num_rows << "-" << this->num_cols << ".bin";
	return ss.str();
}

template<typename T, const uint32_t THREADS_PER_BLOCK>
void SmallSlicedCoo<T, THREADS_PER_BLOCK>::multiply(const T *v, T *r){
//	static const uint32_t NUM_BLOCKS = getNumMultiprocessors();
    bind_x(v);
    dim3 NUM_BLOCK ( min( divideAndCeil(this->num_rows, NUM_ROWS_PER_SLICE) , 65535 ) , divideAndCeil( divideAndCeil(this->num_rows, NUM_ROWS_PER_SLICE) , 65535 ) );
    
	small_sliced_coo_kernel_32<T, NUM_ROWS_PER_SLICE, THREADS_PER_BLOCK> <<< NUM_BLOCK , THREADS_PER_BLOCK>>>(
					this->num_rows,
	                divideAndCeil(this->num_rows, NUM_ROWS_PER_SLICE),
//	                divideAndCeil(this->num_cols, this->GROUP_SIZE),
//	                this->GROUP_SIZE,
	                thrust::raw_pointer_cast(&this->dv_col[0]),
                    thrust::raw_pointer_cast(&this->dv_row[0]),
                    thrust::raw_pointer_cast(&this->dv_value[0]),
	                thrust::raw_pointer_cast(&this->dv_offsets[0]),
	                v,
	                r);
    unbind_x(v);
}
/*
template<>
void SmallSlicedCoo<uint32_t>::multiply(const uint32_t *v, uint32_t *r){
	static const uint32_t NUM_BLOCKS = getNumMultiprocessors();
	small_sliced_coo_kernel_32<uint32_t, THREADS_PER_BLOCK, NUM_ROWS_PER_SLICE> <<<NUM_BLOCKS, THREADS_PER_BLOCK>>>(
					this->num_rows,
	                divideAndCeil(this->num_rows, NUM_ROWS_PER_SLICE),
	                divideAndCeil(this->num_cols, this->GROUP_SIZE),
	                this->GROUP_SIZE,
	                thrust::raw_pointer_cast(&this->dv_coo[0]),
	                thrust::raw_pointer_cast(&this->dv_offsets[0]),
	                v,
	                r);
}*/

}


#endif /* SMALL_SLICED_COO_H_ */
