#include <stdint.h>
#include "util/data_input_stream.h"
#include "util/data_output_stream.h"
#include "matrix/MatrixInput.h"
#include "CutOffFinder.cuh"
#include "util/timer.h"

#include <iostream>
#include <fstream>
#include <algorithm>

using namespace std;
using namespace cuda_bwc;

char *filename;
int blocksize = 32;
int gpu = 0;
bool verbose = false;
bool trial = false;

struct vector_sort { 
    csr_matrix * data;

    vector_sort( csr_matrix *d ) : data(d) {}

    bool operator () ( int i, int j ) {
        
        return ( data->row_offsets[i+1] - data->row_offsets[i] > data->row_offsets[j+1] - data->row_offsets[j]); 
    } 

};


template<typename Double>
void preprocessing ( MMFormatInput &f ) {
    
    uint32_t maxSize = max(f.getNumRows(), f.getNumCols());
    f.perm.resize(maxSize, 0);

    for (int i=0; i<maxSize; i++) f.perm[i] = i;

    vector_sort mysort ( &f.data );
    
    sort( f.perm.begin() , f.perm.begin() + f.getNumRows(), mysort );

//    for (int i=0; i<perm.size(); i++) 
//        cout << perm[i] << " : " << data[perm[i]].size() << endl;     

    if (trial == true) {
        CutOffFinder<Double> cof(f);
        Timer timeex;
        timeex.start();
        cof.execute_trial(verbose);
        cerr << "Spending " << timeex.stop() << "s in finding cutoff" << endl;
        cof.printResult(filename);
        cof.writeCache(filename);
    } 
    else  {
        CutOffFinder<Double> cof( f );
        Timer timeex;
        timeex.start();
        cof.execute(verbose);
        cerr << "Spending " << timeex.stop() << "s in finding cutoff" << endl;
        cof.printResult(filename);
        cof.writeCache(filename);
    }
}

int32_t BalancedMatrix::lastStartRowNo = -1;
int32_t BalancedMatrix::lastNumRows = 0;
std::vector<Slice> BalancedMatrix::vec;
std::priority_queue<Slice*, std::vector<Slice*>, Slice > BalancedMatrix::pq;


void print_usage(char **args) {
    cout << "Usage: prep <mat> [blocksize] [verbose]" << endl;
}

int main(int argc, char **args) {

    int opt;
    if (argc < 2){
        print_usage(args);
        return 0;
    }

    while ((opt = getopt(argc, args, "m:b:vtg:"))!= -1){
        switch (opt){
            case 'b': // block size 
                blocksize = atoi(optarg);
                break;
            case 'm': // matrix file
                filename = optarg;
                break;
            case 'v':
                verbose = true;            
                break;
            case 't':
                trial = true;
                break;
            case 'g':
                gpu = atoi(optarg);    
                break;
            default :
                cerr << "unknown argument: " << (char)opt << endl;
                break;
        }
    }

    cerr << "Using device : " << gpu << endl;
    cudaSetDevice(gpu);

    ifstream inf(filename);

    Timer timer;

    vector<uint32_t> tempperm;
    MMFormatInput f(inf, tempperm);
    
    cerr << "Begin reading matrix ... " << endl;
    f.readIt();

    cerr << "Begin preprocessing ... " << endl;

    timer.start();
    if (blocksize == 32)
        preprocessing<float>(f);    
    else
        preprocessing<double>(f);
    double time = timer.stop();

    cerr << "End preprocessing in " << time << " sec " << endl;

    return 0;
}
